from typing import Any

from flask import Flask, render_template, current_app, Blueprint

client_api = Blueprint('client_api', 'terracotta.client')


@client_api.route('/', methods=['GET'])
def get_map() -> Any:
    return render_template(
        'app.html', hostname=current_app.config['terracotta_hostname']
    )


def create_app(hostname: str) -> Flask:
    client_app = Flask('terracotta.client')
    client_app.config['terracotta_hostname'] = hostname
    client_app.register_blueprint(client_api)
    return client_app



"""cache.py

Custom cache implementations.
"""

from typing import Tuple, Callable, Any

import sys
import zlib

import numpy as np
from cachetools import LFUCache

CompressionTuple = Tuple[bytes, bytes, str, Tuple[int, int]]
SizeFunction = Callable[[CompressionTuple], int]


class CompressedLFUCache(LFUCache):
    """Least-frequently-used cache with ZLIB compression"""

    def __init__(self, maxsize: int, compression_level: int):
        super().__init__(maxsize, self._get_size)
        self.compression_level = compression_level

    def __getitem__(self, key: Any) -> np.ma.MaskedArray:
        compressed_item = super().__getitem__(key)
        return self._decompress_tuple(compressed_item)

    def __setitem__(self, key: Any, value: np.ma.MaskedArray) -> None:
        super().__setitem__(key, self._compress_ma(value))

    def _compress_ma(self,
                     arr: np.ma.MaskedArray) -> CompressionTuple:
        compressed_data = zlib.compress(arr.data, self.compression_level)
        mask_to_int = np.packbits(arr.mask.astype(np.uint8))
        compressed_mask = zlib.compress(mask_to_int, self.compression_level)
        out = (
            compressed_data,
            compressed_mask,
            arr.dtype.name,
            arr.shape
        )
        return out

    def _decompress_tuple(self,
                          compressed_data: CompressionTuple) -> np.ma.MaskedArray:
        data_b, mask_b, dt, ds = compressed_data
        data = np.frombuffer(zlib.decompress(data_b), dtype=dt).reshape(ds)
        mask = np.frombuffer(zlib.decompress(mask_b), dtype=np.uint8)
        mask = np.unpackbits(mask)[:np.prod(ds)]
        mask = mask.reshape(ds)
        return np.ma.masked_array(data, mask=mask)

    @staticmethod
    def _get_size(x: Tuple) -> int:
        sizes = map(sys.getsizeof, x)
        return sum(sizes)



"""scripts/cli.py

Entry point for CLI.
"""

from typing import Any, Mapping
import sys

import click

from terracotta.scripts.click_types import TOMLFile
from terracotta import get_settings, update_settings, logs, __version__


@click.group('terracotta', invoke_without_command=True)
@click.option('-c', '--config', type=TOMLFile(), default=None,
              help='Update global settings from this TOML file.')
@click.option('--loglevel', help='Set level for log messages', default=None,
              type=click.Choice(['debug', 'info', 'warning', 'error', 'critical']))
@click.version_option(version=__version__)
@click.pass_context
def cli(ctx: click.Context,
        config: Mapping[str, Any] = None,
        loglevel: str = None) -> None:
    """The command line interface for the Terracotta tile server.

    All flags must be passed before specifying a subcommand.

    Example:

        $ terracotta -c config.toml connect localhost:5000

    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())

    # update settings from config file
    if config is not None:
        update_settings(**config)

    # setup logging
    settings = get_settings()

    if loglevel is None:
        loglevel = settings.LOGLEVEL

    logs.set_logger(loglevel, catch_warnings=True)


def entrypoint() -> None:
    try:
        cli(obj={})
    except Exception:
        import logging
        logger = logging.getLogger(__name__)
        logger.exception('Uncaught exception!', exc_info=True)
        sys.exit(1)


from terracotta.scripts.connect import connect
cli.add_command(connect)

from terracotta.scripts.ingest import ingest
cli.add_command(ingest)

from terracotta.scripts.optimize_rasters import optimize_rasters
cli.add_command(optimize_rasters)

from terracotta.scripts.serve import serve
cli.add_command(serve)


if __name__ == '__main__':
    entrypoint()



"""server/rgb.py

Flask route to handle /rgb calls.
"""

from typing import Any, Mapping, Dict, Tuple
import json

from marshmallow import Schema, fields, validate, pre_load, ValidationError, EXCLUDE
from flask import request, send_file

from terracotta.server.flask_api import convert_exceptions, TILE_API


class RGBQuerySchema(Schema):
    keys = fields.String(required=True, description='Keys identifying dataset, in order')
    tile_z = fields.Int(required=True, description='Requested zoom level')
    tile_y = fields.Int(required=True, description='y coordinate')
    tile_x = fields.Int(required=True, description='x coordinate')


class RGBOptionSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    r = fields.String(required=True, description='Key value for red band')
    g = fields.String(required=True, description='Key value for green band')
    b = fields.String(required=True, description='Key value for blue band')
    r_range = fields.List(
        fields.Number(allow_none=True), validate=validate.Length(equal=2), example='[0,1]',
        missing=None, description='Stretch range [min, max] to use for red band as JSON array'
    )
    g_range = fields.List(
        fields.Number(allow_none=True), validate=validate.Length(equal=2), example='[0,1]',
        missing=None, description='Stretch range [min, max] to use for green band as JSON array'
    )
    b_range = fields.List(
        fields.Number(allow_none=True), validate=validate.Length(equal=2), example='[0,1]',
        missing=None, description='Stretch range [min, max] to use for blue band as JSON array'
    )
    tile_size = fields.List(
        fields.Integer(), validate=validate.Length(equal=2), example='[256,256]',
        description='Pixel dimensions of the returned PNG image as JSON list.'
    )

    @pre_load
    def process_ranges(self, data: Mapping[str, Any], **kwargs: Any) -> Dict[str, Any]:
        data = dict(data.items())
        for var in ('r_range', 'g_range', 'b_range', 'tile_size'):
            val = data.get(var)
            if val:
                try:
                    data[var] = json.loads(val)
                except json.decoder.JSONDecodeError as exc:
                    raise ValidationError(f'Could not decode value for {var} as JSON') from exc
        return data


@TILE_API.route('/rgb/<int:tile_z>/<int:tile_x>/<int:tile_y>.png', methods=['GET'])
@TILE_API.route('/rgb/<path:keys>/<int:tile_z>/<int:tile_x>/<int:tile_y>.png', methods=['GET'])
@convert_exceptions
def get_rgb(tile_z: int, tile_y: int, tile_x: int, keys: str = '') -> Any:
    """Return the requested RGB tile as a PNG image.
    ---
    get:
        summary: /rgb (tile)
        description: Combine three datasets to RGB image, and return tile as PNG
        parameters:
            - in: path
              schema: RGBQuerySchema
            - in: query
              schema: RGBOptionSchema
        responses:
            200:
                description:
                    PNG image of requested tile
            400:
                description:
                    Invalid query parameters
            404:
                description:
                    No dataset found for given key combination
    """
    tile_xyz = (tile_x, tile_y, tile_z)
    return _get_rgb_image(keys, tile_xyz=tile_xyz)


class RGBPreviewQuerySchema(Schema):
    keys = fields.String(required=True, description='Keys identifying dataset, in order')


@TILE_API.route('/rgb/preview.png', methods=['GET'])
@TILE_API.route('/rgb/<path:keys>/preview.png', methods=['GET'])
@convert_exceptions
def get_rgb_preview(keys: str = '') -> Any:
    """Return the requested RGB dataset preview as a PNG image.
    ---
    get:
        summary: /rgb (preview)
        description: Combine three datasets to RGB image, and return preview as PNG
        parameters:
            - in: path
              schema: RGBPreviewQuerySchema
            - in: query
              schema: RGBOptionSchema
        responses:
            200:
                description:
                    PNG image of requested tile
            400:
                description:
                    Invalid query parameters
            404:
                description:
                    No dataset found for given key combination
    """
    return _get_rgb_image(keys)


def _get_rgb_image(keys: str, tile_xyz: Tuple[int, int, int] = None) -> Any:
    from terracotta.handlers.rgb import rgb

    option_schema = RGBOptionSchema()
    options = option_schema.load(request.args)

    some_keys = [key for key in keys.split('/') if key]

    rgb_values = (options.pop('r'), options.pop('g'), options.pop('b'))
    stretch_ranges = tuple(options.pop(k) for k in ('r_range', 'g_range', 'b_range'))

    image = rgb(
        some_keys, rgb_values, stretch_ranges=stretch_ranges, tile_xyz=tile_xyz, **options
    )

    return send_file(image, mimetype='image/png')



"""drivers/base.py

Base class for drivers.
"""

from typing import Callable, Mapping, Any, Tuple, Sequence, Dict, Union, TypeVar
from abc import ABC, abstractmethod
from collections import OrderedDict
import functools
import contextlib

Number = TypeVar('Number', int, float)
T = TypeVar('T')


def requires_connection(fun: Callable[..., T]) -> Callable[..., T]:
    @functools.wraps(fun)
    def inner(self: Driver, *args: Any, **kwargs: Any) -> T:
        with self.connect():
            return fun(self, *args, **kwargs)
    return inner


class Driver(ABC):
    """Abstract base class for all Terracotta data backends.

    Defines a common interface for all drivers.
    """
    _RESERVED_KEYS = ('limit', 'page')

    db_version: str  #: Terracotta version used to create the database
    key_names: Tuple[str]  #: Names of all keys defined by the database

    @abstractmethod
    def __init__(self, url_or_path: str) -> None:
        self.path = url_or_path

    @abstractmethod
    def create(self, keys: Sequence[str], *,
               key_descriptions: Mapping[str, str] = None) -> None:
        # Create a new, empty database (driver dependent)
        pass

    @abstractmethod
    def connect(self) -> contextlib.AbstractContextManager:
        """Context manager to connect to a given database and clean up on exit.

        This allows you to pool interactions with the database to prevent possibly
        expensive reconnects, or to roll back several interactions if one of them fails.

        Note:

            Make sure to call :meth:`create` on a fresh database before using this method.

        Example:

            >>> import terracotta as tc
            >>> driver = tc.get_driver('tc.sqlite')
            >>> with driver.connect():
            ...     for keys, dataset in datasets.items():
            ...         # connection will be kept open between insert operations
            ...         driver.insert(keys, dataset)

        """
        pass

    @abstractmethod
    def get_keys(self) -> OrderedDict:
        """Get all known keys and their fulltext descriptions.

        Returns:

            An :class:`~collections.OrderedDict` in the form
            ``{key_name: key_description}``

        """
        pass

    @abstractmethod
    def get_datasets(self, where: Mapping[str, str] = None,
                     page: int = 0, limit: int = None) -> Dict[Tuple[str, ...], Any]:
        # Get all known dataset key combinations matching the given constraints,
        # and a handle to retrieve the data (driver dependent)
        pass

    @abstractmethod
    def get_metadata(self, keys: Union[Sequence[str], Mapping[str, str]]) -> Dict[str, Any]:
        """Return all stored metadata for given keys.

        Arguments:

            keys: Keys of the requested dataset. Can either be given as a sequence of key values,
                or as a mapping ``{key_name: key_value}``.

        Returns:

            A :class:`dict` with the values

            - ``range``: global minimum and maximum value in dataset
            - ``bounds``: physical bounds covered by dataset in latitude-longitude projection
            - ``convex_hull``: GeoJSON shape specifying total data coverage in latitude-longitude
              projection
            - ``percentiles``: array of pre-computed percentiles from 1% through 99%
            - ``mean``: global mean
            - ``stdev``: global standard deviation
            - ``metadata``: any additional client-relevant metadata

        """
        pass

    @abstractmethod
    # TODO: add accurate signature if mypy ever supports conditional return types
    def get_raster_tile(self, keys: Union[Sequence[str], Mapping[str, str]], *,
                        tile_bounds: Sequence[float] = None,
                        tile_size: Sequence[int] = (256, 256),
                        preserve_values: bool = False,
                        asynchronous: bool = False) -> Any:
        """Load a raster tile with given keys and bounds.

        Arguments:

            keys: Keys of the requested dataset. Can either be given as a sequence of key values,
                or as a mapping ``{key_name: key_value}``.
            tile_bounds: Physical bounds of the tile to read, in Web Mercator projection (EPSG3857).
                Reads the whole dataset if not given.
            tile_size: Shape of the output array to return. Must be two-dimensional.
                Defaults to :attr:`~terracotta.config.TerracottaSettings.DEFAULT_TILE_SIZE`.
            preserve_values: Whether to preserve exact numerical values (e.g. when reading
                categorical data). Sets all interpolation to nearest neighbor.
            asynchronous: If given, the tile will be read asynchronously in a separate thread.
                This function will return immediately with a :class:`~concurrent.futures.Future`
                that can be used to retrieve the result.

        Returns:

            Requested tile as :class:`~numpy.ma.MaskedArray` of shape ``tile_size`` if
            ``asynchronous=False``, otherwise a :class:`~concurrent.futures.Future` containing
            the result.

        """
        pass

    @staticmethod
    @abstractmethod
    def compute_metadata(data: Any, *,
                         extra_metadata: Any = None,
                         **kwargs: Any) -> Dict[str, Any]:
        # Compute metadata for a given input file (driver dependent)
        pass

    @abstractmethod
    def insert(self, keys: Union[Sequence[str], Mapping[str, str]],
               handle: Any, **kwargs: Any) -> None:
        """Register a new dataset. Used to populate metadata database.

        Arguments:

            keys: Keys of the dataset. Can either be given as a sequence of key values, or
                as a mapping ``{key_name: key_value}``.
            handle: Handle to access dataset (driver dependent).

        """
        pass

    @abstractmethod
    def delete(self, keys: Union[Sequence[str], Mapping[str, str]]) -> None:
        """Remove a dataset from the metadata database.

        Arguments:

            keys:  Keys of the dataset. Can either be given as a sequence of key values, or
                as a mapping ``{key_name: key_value}``.

        """
        pass

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(\'{self.path}\')'



"""scripts/http_utils.py

Various utilities to work with HTTP connections
"""

from typing import Optional, Sequence


def check_socket(host: str, port: int) -> bool:
    """Check if given port can be listened to"""
    import socket
    from contextlib import closing

    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(2)
        try:
            sock.bind((host, port))
            sock.listen(1)
            return True
        except socket.error:
            return False


def find_open_port(port_range: Sequence[int]) -> Optional[int]:
    """Return first open port in port_range, or None if no open port is found"""
    for port_candidate in port_range:
        if check_socket('localhost', port_candidate):
            return port_candidate

    return None



"""handlers/datasets.py

Handle /datasets API endpoint.
"""

from typing import Mapping, List  # noqa: F401
from collections import OrderedDict

from terracotta import get_settings, get_driver
from terracotta.profile import trace


@trace('datasets_handler')
def datasets(some_keys: Mapping[str, str] = None,
             page: int = 0, limit: int = 500) -> 'List[OrderedDict[str, str]]':
    """List all available key combinations"""
    settings = get_settings()
    driver = get_driver(settings.DRIVER_PATH, provider=settings.DRIVER_PROVIDER)

    with driver.connect():
        dataset_keys = driver.get_datasets(
            where=some_keys, page=page, limit=limit
        ).keys()
        key_names = driver.key_names

    return [OrderedDict(zip(key_names, ds_keys)) for ds_keys in dataset_keys]
